/*
 *
 * INNOVA software is coded by vecna@s0ftpj.org, history about develop
 * should be find on HISTORY file, license on the LICENSE file,
 * generic information on the README file, those, and other files
 * are included on the innova package, should be downloaded from
 * www.s0ftpj.org
 *
 * for one italian description you should read README.it
 *
 * 0.0.1:		base innova core, using iptables only
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <dlfcn.h>
#include <signal.h>
#include <sys/param.h>
#include <string.h>

#include "innova.h"
#include "userdefs.h"

/* 
 * used by the signal handler for reach modules function, kernel
 * and task_struct current have been contaminated my coding style ..
 */
struct innova_struct *current;

/*
 * this function is used because before any hook should be called 
 * the foo function related to the mangle suite of expandible functions
 */
static inline int 
call_mangle_function(struct innova_struct *is, struct innova_packet *pkt, unsigned char hook)
{
	int ret =0x00;

	if(is->m->foo_mangle_f !=NULL)
		if((ret =is->m->foo_mangle_f(hook, is, pkt, is->m->foo)) < 0)
			return ret;
	
	switch(hook)
	{
		case INIT_HOOK:
			ret =is->m->mangle_init(is);
			break;
		case CLEANUP_HOOK:
			ret =is->m->mangle_cleanup(is, errno);
			break;
		case LOCAL_M_HOOK:
			ret =is->m->local_mangle(is, pkt);
			break;
		case REMOTE_M_HOOK:
			ret =is->m->remote_mangle(is, pkt);
			break;
		case UNKNOW_M_HOOK:
			/* only managed from foo_mangle_f if exist */
			break;
		case TIMEEXC_HOOK:
			ret =is->m->io_timeexceed(is, &is->o->timeout);
			break;
		default:
			innova_p(FATAL, "invalid use of call_mangle_function on your code!\n");
	}

	return ret;
}

static inline int handle_packets(struct innova_struct *is)
{
	int nfds =0x00;
	struct innova_packet *pkt;

	/* 
	 * ret is used for check return value and is returned with 
	 * error for be passed to the cleanup function
	 */
	int ret =0x00;
	/* timeout in milliseconds */
	int ms_timeout;

	if(!is->o->timeout)
		ms_timeout =DEFAULT_TIMEOUT;
	else
		ms_timeout =is->o->timeout;

	while((pkt =wait_packet(ms_timeout)) !=(void *)-1)
	{
		/* wait_packet return null when timeout is expired */
		if(pkt ==NULL)
		{
			if((ret =call_mangle_function(is, pkt, TIMEEXC_HOOK)) < 0)
				return ret;

			continue;
		}

		/*
		 * here is undestand if the packet is coming from netlink device, is
		 * generated from local application and must be send for local_mangle
		 * API in the plugin or is readed from datalink layer, and passed to 
		 * remote_mangle API.
		 *
		 * should be marked as UNKNOW, as extension usable with the foo function
		 * pointer present on innova_struct and innova_packet for possibile 
		 * extensions.
		 *
		 * return value depend from the plugins function, local_mangle and 
		 * remote_mangle. < 0 is error, 0 is definied PACKET_OK, 1 is
		 * PACKET_DROP.
		 */
		switch(pkt->IO)
		{
			case LOCAL:
				if((ret =call_mangle_function(is, pkt, LOCAL_M_HOOK)) < 0)
					return ret;
				break;
			case REMOTE:
				if((ret =call_mangle_function(is, pkt, REMOTE_M_HOOK)) < 0)
					return ret;
				break;
			default:
				if((ret =call_mangle_function(is, pkt, UNKNOW_M_HOOK)) < 0)
					return ret;
		}

		if(ret ==PACKET_DROP)
			continue;

		if((ret =reinject_packet(pkt, is)) < 0)
			return ret;
	}

	/* 
	 * is function reach this point, is not returned and poll have
	 * been give an error
	 */
	return innova_p(FATAL_ERRNO, "polling the %d requested stream", nfds);
}

void signal_handler(int signum)
	{
	static int siglock;

	/*
	 * siglock is used if segmentation fault is raised from
	 * plugin, should be raised also calling the CLEANUP_HOOK
	 * and should be created a loop, after the first FATAL
	 * signal handling, siglock is set and SIGKILL is used if 
	 * loop is detected 
	 */
	if(siglock)
		{
		fprintf(stderr, "double kill!, received signal %d"
				" on cleaning innova... ", signum);
		raise(SIGKILL);
		}

	siglock =1;

	/*
	 * with errno set to 0, I mean user request for close 
	 * innova, not an error
	 */
	switch(signum)
	{
		case SIGTERM:
		case SIGINT:
		case SIGPIPE:
			errno =0x00;
			break;
		case SIGILL:
		case SIGSEGV:
			errno =0xffff;
		default:
			break;
	}

	innova_p(ERROR, "forced quit for signal: %d", signum);

	/* calling option cleanup and networking cleanup */
	call_mangle_function(current, NULL, CLEANUP_HOOK);
	netlink_cleanup();

	/* should happens that signal is delivered before network init */
	if(current->MTU)
		mtu_restore(current, current->MTU);

	signal(SIGTERM, SIG_DFL);
	raise(SIGTERM);
	}

void signal_handling(void)
	{
	int i; 
	int siglist[]=
		{
		SIGHUP, SIGINT, SIGQUIT, SIGILL, SIGABRT,
		SIGFPE, SIGSEGV, SIGPIPE, SIGTERM, SIGTTIN, 
		SIGTTOU, SIGBUS, SIGPROF, SIGSYS, SIGTRAP, 
		SIGVTALRM, SIGXCPU, SIGXFSZ, 
		SIGIOT, SIGSTKFLT, 
		};

	for(i=0x00; siglist[i] !=0x00; i++)
		signal(siglist[i], signal_handler);
	}

static int check_plugin_support(int pv)
{
	if(pv < PLUGIN_FORMAT)
		return 1;
	else
		return 0;
}

const char *spam=
"innova is a sw with a very troubled history, vecna@s0ftpj.org " \
"aka Claudio Agosti \nhave been coded the first published release " \
"(but the third personal release :) \nmore or less on the september 2003." \
"\non http://www.s0ftpj.org/innova you should find other information"
"\n\n";


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                       *
 *                                                       *
 |        ___________                                    *
 |       ( innova !! )                                   |
 ^        -----------                                    |
        	 o   ^__^                                |
	          o  (oo)\_______                        $
	             (__)\       )\/\
	        	 ||----w |
		         ||     ||
 *
 *     THREE MAN MAKE A TIGER, ONE COW MAKE A THOUGHT     
 * 
 * 
 */

int main(const int argc, const char **argv)
{
	static struct innova_struct is;
	static struct innova_options iopt;
	static struct innova_mangle io;
	void *opt_h;
	char plugin_path[MAXPATHLEN], *pdir;
	int pv, ret =0;

	if(argc < 2)
		innova_p(FATAL, "%sfor use innova, I suggest \"%s -h\" ", spam, argv[0]);

	/*
	 * globalize innnova_struct, fill is->o (user option)
	 * and is->m (mangle function retrivered from plugin)
	 */
	is.o =&iopt;
	is.m =&io;
	current =&is;

	parse_opt(&iopt, argc, argv);

	pdir = iopt.plug_dir != NULL ? iopt.plug_dir : getenv("PWD");

	snprintf(plugin_path, MAXPATHLEN, "%s/%s.so", pdir, iopt.plug_name);
	
	if((opt_h =dlopen(plugin_path, 0x01)) ==NULL)
		innova_p(FATAL , "mangle module %s", dlerror());

	/* informational function */
	read_optsym((void *)&io.get_io_desc, opt_h, "get_io_desc", plugin_path);

	/* initialization, cleanup and options parsing */
	read_optsym((void *)&io.mangle_init, opt_h, "mangle_init", plugin_path);
	read_optsym((void *)&io.mangle_cleanup, opt_h, "mangle_cleanup", plugin_path);

	read_optsym((void *)&io.local_mangle, opt_h, "local_mangle", plugin_path);
	read_optsym((void *)&io.remote_mangle, opt_h, "remote_mangle", plugin_path);

	/* when timeout (user defined or 1 second default) expired, is called */
	read_optsym((void *)&io.io_timeexceed, opt_h, "io_timeexceed", plugin_path);

	/* function for parse options plugin related */
	read_optsym((void *)&io.mangle_opt_parse, opt_h, "option_parser", plugin_path);

	innova_p(INFO, "innova start init of plugin %s: %s", plugin_path, io.get_io_desc(&pv));

	if(check_plugin_support(pv))
	{
		innova_p(FATAL, "plugin [%s] is version %d, innova [%s] required plugin version %d",
			iopt.plug_name,
			pv,
			INNOVA_RELEASE,
			PLUGIN_FORMAT
		);
	}

	signal_handling();

	/* fixing MTU for possible decrasing */
	is.min_MSS =is.MTU =mtu_init(&is, &iopt);

	/*
	 * before mangle initilization, is passed the related string with 
	 * the plugins options, innova_struct is and innova_options iopt
	 * is passed, usually for fill static field on plugin object.
	 *
	 * module receive a char **argv, argc more or less how our shell
	 * format argument for main(), splitting of iopt->plug_opt is 
	 * made from format_options (called after getopt "-o".
	 *
	 * here should be add remote_match_ext and local_match_ext, other
	 * than apply some change to the remote, protocol, server and 
	 * other local/remote network discriminator, who should be 
	 * changed with plugin's configuration file.
	 *
	 * if return 0 is ok, if return a value this mean error.
	 */
	if(io.mangle_opt_parse(&is, &iopt))
		innova_p(FATAL, 
			"%s receives error on reading option from \"-o\" switch", 
			io.get_io_desc(&pv)
		);
	
	/* 
	 * initialization of networking iterface
	 */
	netlink_init(&is, &iopt);

	/* ipulog_handle and 0x00 */
	add_innova_stream(is.local, 0x00);

	/* NULL as ipulog_handle and valid file descriptor */
	add_innova_stream(NULL, is.lremotp);

	/*
	 * this is innova main loop, simple here is called init routine and
	 * the "handle_packets" function, (where packets is received and is
	 * manageg timeout function, reading, mangling and resending) whenever
	 * one of this function generate an error (shoul be a disconnected peer
	 * or deleted file or so other ...) handle_packets return the negative 
	 * errno and this situation is managed from the cleanup function,
	 * cleanup function should manage the error and make a decision about
	 * him return value, if is returned a negative value innova exit from
	 * him main loop, or should be ignored (returing > 0 value) or should 
	 * be resterted innova session calling a new istance of mangle_init
	 * (this happens if 0 is returned) 
	 */
	while(1) 
	{
		/* 
		 * before any initialization, is called foo_options extension,
		 * for optionally options rebuilding.
		 *
		 * ret should be used for check error code, 
		 * argv is cleaned 
		 */
		if(iopt.foo_options_f != NULL)
			iopt.foo_options_f(ret, (char **)argv, (void *)&iopt);

		/* mangle init is called for start mangle session */
		ret =call_mangle_function(current, NULL, INIT_HOOK);

		/* 
		 * -ERROR break cicle, 0 return to the head of cicle, > 0 is
		 * correct value and mangle session should continue
		 */
		if(ret < 0)
			break;
		else if(!ret)
			continue;
		else
			;

		/*
		 * handle_packets is the main loop where innova run for read/mangle/inject
		 * packets, negative return value mean end of session.
		 */
		if((ret =handle_packets(&is)) < 0)
		{
			ret =call_mangle_function(current, NULL, CLEANUP_HOOK);

			if(ret < 0)
				break;
			else 
				continue;
		}
	}

	/* 
	 * if you want a specific error is print at end of excution, change the positive
	 * value of return code in CLEANUP_HOOK, innova_p resolve him string value.
	 */
	errno =ret;

	innova_p(ERROR_ERRNO, "fatal error occurred", errno);
	netlink_cleanup();

	return ret;
}
