/*
 *
 * INNOVA software is coded by vecna@s0ftpj.org, history about develop
 * should be find on HISTORY file, license on the LICENSE file,
 * generic information on the README file, those, and other files
 * are included on the innova package, should be downloaded from
 * www.s0ftpj.org
 *
 * for one italian description you should read README.it
 *
 */

#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>

#include <net/route.h>
#include <net/if.h>
#include <arpa/inet.h>

#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/param.h>

#include <netdb.h>

#include "l/libipulog.h"
#include "userdefs.h"

#define INNOVA_VERSION 	"0.0.1"
#define PLUGIN_FORMAT	1
#define INNOVA_RELEASE 	"innova-0.0.1"
#define INNOVA_HOMEPAGE "http://www.s0ftpj.org/innova"
#define INNOVA_AUTHOR 	"Claudio Agosti <vecna@s0ftpj.org>"


#define v_innova_struct_p void *
#define v_innova_packet_p void *
#define v_innova_mangle_p void *
#define v_innova_options_p void *

/*
 * innova options structure
 */
struct innova_options
	{
	char *plug_dir;
	char *plug_name;
	char *plug_opt;	/* passed with -o "options args ..." */
	char *chroot;
	char *interface;

	/* setted remote host, reply remote host */
	struct hostent *remote, *contact;
	/* setted network, reply reply network */
	struct netent *network, *retwork;
	/* remote service discriminated, local service disc.. */
	struct servent *rem_serv, *loc_serv;
	/* local protocol reader, remote protocol waited */
	struct protoent *protocol, *retocol;

	char *local_match_ext;
	char *remote_match_ext;
	/* presence of match extension override protocol and server specific */
	
	int timeout;

	char **argv;	/* pointing to plug_opt */
	int argc;

	/* should be used for some expansion ... */
	int (*foo_options_f)(int, char **, v_innova_options_p);
	void *foo;
	};

/* return value from plugin options */
#define PACKET_OK	0
#define PACKET_DROP	1
/* and < 0 is error code */

struct innova_mangle
	{
	char *(* get_io_desc)(int *);

#define INIT_HOOK	0
	int (* mangle_init)(v_innova_struct_p);

#define CLEANUP_HOOK	1
	int (* mangle_cleanup)(v_innova_struct_p, int);

#define OPT_PARSE_HOOK	2
	int (* mangle_opt_parse)(v_innova_struct_p, v_innova_options_p);

#define LOCAL_M_HOOK	3
#define REMOTE_M_HOOK	4
#define UNKNOW_M_HOOK	5
	int (* local_mangle)(v_innova_struct_p, v_innova_packet_p);
	int (* remote_mangle)(v_innova_struct_p, v_innova_packet_p);

#define TIMEEXC_HOOK	6
	int (* io_timeexceed)(v_innova_struct_p, int *);

	/* 
	 * should be used for some expansion, il called always
	 * before any *_mangle function
	 */
	int (* foo_mangle_f)(unsigned char, v_innova_struct_p, v_innova_packet_p, void *);
	void *foo;
	};

/* 
 * innova_packet struct describe the packets fragmentated for is 
 * required to the engine that must resent packets, for build 
 * checksum and for reorder header on the packets, this struct 
 * describe the packets on innova code. is the struct for give 
 * packets to the manipoulation function.
 */
struct innova_packet
	{
	/* ip header and ip options */
	struct iphdr *ip;
	unsigned char *ip_opt;
	unsigned char ip_opt_len;

	/* tcp header and tcp options */
	struct tcphdr *tcp;
	unsigned char *tcp_opt;
	unsigned char tcp_opt_len;

	/* udp header */
	struct udphdr *udp;

	/* icmp header */
	struct icmphdr *icmp;

	/* raw header usable on datalink socket */
	void *raw;
	int length;

	/* 
	 * data, used from all kind of packet, for received
	 * user data (tcp connect/socket unix) or data field
	 * on raw packet
	 */
	unsigned char *data;
	int data_len;

	/* 
	 * for undestound if packet is coming from local socket or
	 * from remote 
	 */
#define LOCAL	0
#define REMOTE	1
	unsigned char IO;

	/* information from netlink header */
	unsigned int sec;
	unsigned int usec;
	unsigned int mark;
	unsigned int diff;

	/* for plugin who needed to generate a chain of packets */
	struct innova_packet *next;
	
	/* should be used for some expansion */
	int (* foo_packet_f)(v_innova_struct_p, v_innova_packet_p);
	void *foo;
	};

struct innova_struct
	{
	struct ipulog_handle *local;
	int lremotp;	/* listener for remote packets */
	int rawfd;

	struct innova_mangle *m;
	struct innova_options *o;

	int local_nlg;	/* netlink group for local packets */
	int local_mark; /* mark for local packet to drop */
	int innova_table; /* table for iproute2 policy routing */

	int MTU, min_MSS;
			/* 
			 * for manage tcp connection we need to
			 * track MSS requested on tcp connection
			 * and lowered out MTU, this is lazy hack
			 */
	int ifd;	/* interface file descriptor, for open them only
			 * one time 
			 */

	void *foo;
	};

/*
 * utilf.c macro and prototypes
 */

#define BUFSIZE 	4096
/* 
 * this dimension is used for a shared static buffer, called "tmp_buf", for 
 * mangage packets without dynamic allocation, simple is used from the
 * reading routing to the resending routine, more faster than use two
 * addictional system call (malloc and free are two brk(2) ) each 
 * packet...
 */
inline int append_in_buffer(void *, int);
inline unsigned char *get_b_addr(void);
 /* 
  * append in buffer work over them, and get_b_addr return him address,
  * read the code for undestound him function, on utilf.c there are
  * a lot of comments
  */

/* for get symbol from plugins */
void read_optsym(void *(*), void *, char *, char *);

/* for resolve remote service, network, host and protocol */
struct servent *service_resolve(const char *, char *);
struct hostent *hostname_resolve(const char *);
struct netent *network_resolve(const char *);
struct protoent *protocol_resolve(const char *);

/* for compute ip/tcp/udp/icmp check sum */
inline unsigned int half_cksum(const unsigned short *, int);
inline unsigned short compute_sum(unsigned int);

/* for safe allocation, simple wrapper to malloc + memset */
inline void imalloc(void **, unsigned int);

#define FATAL_ERRNO     1
#define FATAL           2
#define ERROR_ERRNO	3
#define ERROR		4
#define WARN_ERRNO      5
#define WARN            6
#define DEBUG_ERRNO	7
#define DEBUG		8

#define INFO		10
#define LOG		11
int innova_p(char, char *, ...);

/*
 * options.c
 */
void parse_opt(struct innova_options *, int, const char **);
void format_options(struct innova_options *);

/* poll.c */
struct ul_handle_fd
{
	void *ip_handle;
	int fd;
};

#define MAX_PLUG_FDS	32
void add_fd(int, int (*)(struct innova_struct *, int));
void remove_fd(int);
void add_innova_stream(struct ipulog_handle *, int);

#define IOMULTIPLEX_ERROR	(struct innova_packet *) -1
struct innova_packet *wait_packet(int);

/*
 * netlink.c
 */
#define MAXCOMMANDLEN   255
#define DEFAULT_TRASH_INTERFACE "lo"
struct pkt_trace
{
	unsigned short id;
	unsigned short tot_len;
	unsigned int l4_value;
};

/* FIXME - split netlink.c on netio.c e netutils.c */
int netlink_init(struct innova_struct *, struct innova_options *);
void netlink_cleanup(void);
inline struct innova_packet *get_ulog_packet(struct ipulog_handle *, unsigned char *);
inline struct innova_packet *get_datalink_packet(int, unsigned char *);
inline int reinject_packet(struct innova_packet *, struct innova_struct *);
inline int fill_pkt_pointer(struct innova_packet *, unsigned char *);
inline int dump_in_buffer(struct innova_packet *);
void fill_portopt(int, unsigned char, char *, int, struct servent *, struct servent *);

#define IPTCMD_DELETE_RULES(string)                     \
{							\
	char *ipt_action;				\
        if((ipt_action =strchr(string, 'A')) !=NULL)    \
        {                                               \
                *ipt_action ='D';                       \
                system(string);                         \
        }						\
}

void checked_execution(char *);

inline void mtu_tuning(struct innova_packet *);
unsigned int mtu_change(struct innova_struct *, short, short);
unsigned int mtu_init(struct innova_struct *, struct innova_options *);
void mtu_restore(struct innova_struct *, unsigned int);

/*
 * protocol.c function
 */
inline void fix_ip_checksum(struct innova_packet *);
inline void fix_tcp_checksum(struct innova_packet *);
inline void fix_icmp_checksum(struct innova_packet *);
inline void fix_udp_checksum(struct innova_packet *);

int is_supported(char *);
int verify_ip_check(struct iphdr *);

void fill_tcphdr(struct innova_packet *, const unsigned char *, int *);
void fill_udphdr(struct innova_packet *, const unsigned char *, int *);
void fill_icmphdr(struct innova_packet *, const unsigned char *, int *);
void fill_ip_options(struct innova_packet *, const unsigned char *, int *);
