/*
 * file for parsing options - innova project
 * http://www.s0ftpj.org
 *
 * version 0.0.1:	base options support
 */

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>

#include "innova.h"
#include "userdefs.h"

static void print_version(void)
{
	printf( " innova version\t %s\n"
		" release date\t %s\n"
		" main site\t %s\n"
		" author\t\t %s\n",
	INNOVA_VERSION, INNOVA_RELEASE, INNOVA_HOMEPAGE, INNOVA_AUTHOR);

	exit(0);
}

static void print_help(void)
{
	printf(	"\nremote host, plugin and interface is required\n"
		"-r\n-l = remote and local service \n     otherwise rules is "
		"applied to all traffic between local and specified "
		"remote host\n"
		"-i = interface\n     interface for wait packets reply\n"
		"-t = timeout\n     default %d seconds, read about "
		"plugin, should be ignored\n"
		"-d = plugins dir\n     otherwise is used default %s\n"
		"-c = chroot\n     directory for jail innova after "
		"plugin initialization\n"
		"-m = plugin\n     name of plugin\n"
		"-p = protocol\n     protocol selection\n"
		"-o = options\n     option string for plugin\n"
		"-v = version and some info\n"
		"as last value is required hostname or hostname/netmask\n\n",
		DEFAULT_TIMEOUT, INNOVA_P_PATH); 

	exit(0);
}

void parse_opt(struct innova_options *iopt, int argc, const char **argv)
{
	int opt;
	char *net;

	extern char *optarg;
	extern int optind;

	if(argc == 2 && !strcmp(argv[1], "-v"))
		print_version();
	
	if(argc < 6 || !strcmp(argv[1], "-h"))
		print_help();

	/* 
	 * is required resolve hostname before options, check if 
	 * there are simple hostname or hostname/netmask 
	 */
	if((net =strchr(argv[argc -1], '/')) !=NULL)
	{
		/* (*net) is on "/" */
		net++;
		iopt->network =network_resolve(net);

		*(net -1)=0x00;
	}

	iopt->remote =hostname_resolve(argv[argc -1]);

	argc--;

	/*
	 * -r = remote service
	 * -l = local service
	 * -t = timeout
	 * -d = plugins dir
	 * -c = chroot
	 * -m = plugin
	 * -o = options for plugin
	 * -p = protocol
	 * -i = interface
	 */
	while((opt =getopt(argc, (char **)argv, "r:l:t:d:c:i:p:m:o:h")) != -1) 
	{
		switch(opt)
		{
			case 'r':
				{
				char *protoname =NULL;

				if(iopt->protocol !=NULL && iopt->protocol->p_name !=NULL)
					protoname =iopt->protocol->p_name;

				iopt->rem_serv =service_resolve(optarg, protoname);
				break;
				}
			case 'l':
				{
				char *protoname =NULL;

				if(iopt->protocol !=NULL && iopt->protocol->p_name !=NULL)
					protoname =iopt->protocol->p_name;

				iopt->loc_serv =service_resolve(optarg, protoname);
				break;
				}
			case 't':
				iopt->timeout =atoi(optarg);
				break;
			case 'i':
				iopt->interface =strdup(optarg);
				break;
			case 'd':
				iopt->plug_dir =strdup(optarg);
				break;
			case 'c':
				iopt->chroot =strdup(optarg);
				break;
			case 'p':
				iopt->protocol =protocol_resolve(optarg);
				break;
			case 'm':
				iopt->plug_name =strdup(optarg);
				break;
			case 'o':
				iopt->plug_opt =strdup(optarg);
				format_options(iopt);
				break;
			case 'h':
				print_help();
				break;
			default:
				innova_p(FATAL, "invalid options %c (%s)", optind, optarg);
		}

		/* remote discriminator is the some, your plugin on init time should change */
		iopt->contact =iopt->remote;
		iopt->retwork =iopt->network;
		iopt->retocol =iopt->protocol;
	}

	if(iopt->plug_name ==NULL || iopt->interface ==NULL)
		innova_p(FATAL, "bad options: plugin and interface are REQUIRED");
}

/*
 * this function must fill io->argc and io->argv 
 * indexing the options for the plugin, readed on 
 * parse_opt
 */
void format_options(struct innova_options *io)
{
	char *entr, *p;
	int i;

	if((entr =io->plug_opt) ==NULL)
		return;

	/* split starting space */
	while(*entr && *entr ==' ')
		entr++;

	/* string analysis, "arg ant cip la bla" */
	for(i =0; i < strlen(entr); i++)
	{
		if(entr[i] !=' ' && entr[i] !='\\')
			continue;

		if(entr[i] ==' ' && entr[i -1] !='\\')
			io->argc++;
	}

	if(!io->argc && strlen(entr) > 0)
		io->argc =1;

	imalloc((void **)&io->argv, sizeof(char *) * io->argc);

	for(i =0, p =entr; *p ; )
	{
		p =strchr(entr, ' ');

		if(p ==NULL)
		{
			io->argv[i] =strdup(entr);
			break;
		}

		if(*(p -1) != '\\')
		{
			*p =0x0;

			io->argv[i++] =strdup(entr);
			*p =' ';
			entr =++p;
		}
	}
}
