/*
 * innova plugin, coded by Claudio Agosti vecna@s0ftpj.org
 *
 * Mon Oct 13 22:08:31 2003, finished Thu Jan 29 18:35:13 2004
 * (I like lost my time)
 *
 * ip_steganography work to implement transparent steganography 
 * on ip packets, using ip id field.
 *
 * some operating system and other with special patch implement
 * random ip id generation. this can be a nice way to 
 * hide data inside connection, using innova plugin this can
 * be applied transparently with your common client sessions.
 *
 * http://www.gnupg.org
 * http://www.gzip.org
 * http://mcrypt.sf.net
 *
 * these software should be used to crypt and compress file before
 * sending with this plugin. the options passed to this plugin must
 * be "/dump_directory /input_directory"
 * the dump_directory could be filled with ip.ip.ip.ip-port dump
 * of incoming session, on the input_directory it looks for the
 * file with ip.ip.ip.ip try to connect, if founded, is used as
 * input to take the couple of byte to put on ip->id field.
 *
 * (http://www.s0ftpj.org :) for better information you should
 * search information on http://claudio.itapac.net
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>

#include "innova.h"

static char *plug_desc="ip steganography";

static char *dump_path, *source_path;

#define MAXSEQ		4
struct stgcouple_track
{
	unsigned int addr, counter;
	unsigned short port;
	FILE *dump;
	char *fname;
	unsigned int last_seq[MAXSEQ];
};
/* maximun of incoming connection tracked */
#define MAXTRAKSEX	20

static struct stgcouple_track incoming_track[MAXTRAKSEX];
static struct stgcouple_track outgoing_track[MAXTRAKSEX];

#define INCOMING	1
#define OUTGOING	2

char *get_io_desc(int *plugin_version)
{
	*plugin_version =PLUGIN_FORMAT;

	return plug_desc;
}

int mangle_init(struct innova_struct *is)
{
	/*
	 * return < 0 is error, and innova is break,
	 * return 0 is for repeat mangle_init,
	 * return > 0 is for init success, innova continue happy
	 */
	return 1;
}

int mangle_cleanup(struct innova_struct *is, int error)
{
	int i;

	printf("forced closing...\n");

	for(i =0; i < MAXTRAKSEX; i++)
	{
		if(incoming_track[i].dump !=NULL)
		{
			printf("incoming file %s, %d packet\n", 
				incoming_track[i].fname, 
				incoming_track[i].counter
			);
			fclose(incoming_track[i].dump);
			incoming_track[i].dump =NULL;
		}
	}

	for(i =0; i < MAXTRAKSEX; i++)
	{
		if(outgoing_track[i].dump && !feof(outgoing_track[i].dump))
		{
			printf("outgoing opened session %s, "
			       "ending without be finish (%d byte sent)\n",
				outgoing_track[i].fname, 
				outgoing_track[i].counter
			);
			fclose(outgoing_track[i].dump);
			incoming_track[i].dump =NULL;
		}
	}

	return 1;
}

static inline struct stgcouple_track *
get_session(int who, struct innova_packet *pkt)
{
	unsigned int i;
	struct stgcouple_track *list;

	if(who ==INCOMING)
		list =incoming_track;
	else
		list =outgoing_track;

	for(i =0; i < MAXTRAKSEX; i++)
	{
		if(who ==INCOMING)
			if(pkt->ip->saddr ==list[i].addr && 
			   pkt->tcp->source ==list[i].port)
				return &list[i];

		if(who ==OUTGOING)
			if(pkt->ip->daddr ==list[i].addr && 
			   pkt->tcp->dest ==list[i].port)
				return &list[i];
	}
	return NULL;
}

static inline struct stgcouple_track *
get_next_free(int who)
{
	unsigned int i;
	struct stgcouple_track *list;

	if(who ==INCOMING)
		list =incoming_track;
	else
		list =outgoing_track;

	for(i =0; i < MAXTRAKSEX; i++)
		if(list[i].dump ==NULL)
			return &list[i];

	return NULL;
}

#define	DUPSEQ	0
#define NEWSEQ	1

/* is used xor because seq and ack_seq should not change both */
static inline int check_seq(unsigned int *seqlist, unsigned int last_xor)
{
	unsigned int i;

	for(i =0; i < MAXSEQ; i++)
	{
		if(seqlist[i] ==last_xor)
			return DUPSEQ;
	}

	for(i =(MAXSEQ -1); i > 0; i--)
		seqlist[i] =seqlist[i -1];

	seqlist[0] =last_xor;

	return NEWSEQ;
}

int local_mangle(struct innova_struct *is, struct innova_packet *pkt)
{
	struct stgcouple_track *tracking =get_session(OUTGOING, pkt);

	if(tracking ==NULL && pkt->tcp->syn)
	{
		char fname[MAXPATHLEN], i;

		if((tracking =get_next_free(OUTGOING)) ==NULL)
		{
			printf("outgoing session overcoming limit of %d\n", 
				MAXTRAKSEX);
			return PACKET_OK;
		}

		sprintf(fname, "%s/%s", source_path, 
			inet_ntoa(*(struct in_addr *)&pkt->ip->daddr));

		if((tracking->dump =fopen(fname, "r")) ==NULL)
			return PACKET_OK;

		printf("opened [%s] file for be sent\n", fname);

		tracking->fname =strdup(fname);

		tracking->addr =pkt->ip->daddr;
		tracking->port =pkt->tcp->dest;

		for(i =0; i < MAXSEQ; i++)
			tracking->last_seq[(int)i] =0x00;
	}

	if(tracking ==NULL)
		return PACKET_OK;

	if(pkt->tcp->fin || pkt->tcp->rst)
	{
		printf("closing session [%s] packets sent %d\n", 
			tracking->fname, tracking->counter);

		if(!feof(tracking->dump))
		{
			printf("ERROR! file %s not totally sent, "
			       "%d byte only\n", 
				tracking->fname, tracking->counter *2
			);
		}

		fclose(tracking->dump);
		tracking->dump =NULL;
		tracking->addr = tracking->port =0;

		return PACKET_OK;
	}

	/* to avoid duplicated packet */
	if((check_seq(tracking->last_seq, 
			pkt->tcp->ack_seq ^ pkt->tcp->seq)) ==DUPSEQ)
		return PACKET_OK;

	if(feof(tracking->dump))
	{
		printf("session %s sent after %d byte\n", 
			tracking->fname,
			tracking->counter
		);
		fclose(tracking->dump);
		tracking->dump =NULL;
		tracking->addr = tracking->port =0;
	}
	else
	{
		fread(&pkt->ip->id, 1, sizeof(unsigned short), tracking->dump);
		tracking->counter++;
	}

	return PACKET_OK;
}

int remote_mangle(struct innova_struct *is, struct innova_packet *pkt)
{
	struct stgcouple_track *tracking =get_session(INCOMING, pkt);

	if(tracking==NULL && pkt->tcp->syn)
	{
		char fname[MAXPATHLEN], i;

		if((tracking =get_next_free(INCOMING)) ==NULL)
		{
			printf("ingoing session overcoming limit of %d\n", 
				MAXTRAKSEX);
			return PACKET_OK;
		}

		/* 
		 * YES, THIS IS A BUFFER OVERFLOW!,
		 * root could became root.
		 */
		sprintf(fname, "%s/%s-%d", dump_path, 
			inet_ntoa(*(struct in_addr *)&pkt->ip->saddr),
			htons(pkt->tcp->source)
		);
		
		if((tracking->dump =fopen(fname, "w+")) ==NULL)
		{
			printf("unable to open file %s!!\n", fname);

			return PACKET_OK;
		}
		tracking->fname =strdup(fname);

		tracking->addr =pkt->ip->saddr;
		tracking->port =pkt->tcp->source;

		for(i =0; i < MAXSEQ; i++)
			tracking->last_seq[(int)i] =0x00;
		tracking->counter =0;
	}

	/* untracked session */
	if(tracking ==NULL)
		return PACKET_OK;

	/* to avoid duplicated packet */
	if((check_seq(tracking->last_seq, 
			pkt->tcp->ack_seq ^ pkt->tcp->seq)) ==DUPSEQ)
		return PACKET_OK;
	/* 
	 * sorry for stressing your scheduler calling a lot of system call,
	 * but this is only a demonstration... static connection table,
	 * check only the last sequence number, and other non-performantic
	 * things :)
	 */
	if(pkt->tcp->fin || pkt->tcp->rst)
	{
		printf("closing session [%s] packets received %d\n", 
			tracking->fname, tracking->counter);

		fclose(tracking->dump);
		tracking->dump =NULL;
		tracking->addr = tracking->port =0;
	}
	else
	{
		fwrite(&pkt->ip->id, 1, sizeof(short), tracking->dump);
		tracking->counter++;
	}


	return PACKET_OK;
}

int io_timeexceed(struct innova_struct *is, int *timeout)
{
	return 0x00;
}

void stegano_print_help(void)
{
	fprintf(stderr,
		"steganography plugin simple required two path, one for "
		"incoming sessions dump\nthe other for search data to put"
		" inside the packets\nfirst: dump, eg /tmp\nsecond: input,"
		" eg /encrypted/ (with.any.ip.file) like "
		"/encrypted/192.168.0.1\n"
	);
}

static int check_path(char *path)
{
	struct stat st;

	if(path ==NULL)
		return 0;

	if((stat(path, &st)) ==-1)
		return 0;

	return S_ISDIR(st.st_mode);
}

int option_parser(struct innova_struct *is, struct innova_options *iopt)
{
	if(iopt->plug_opt ==NULL || !strcmp(iopt->plug_opt, "help"))
	{
		stegano_print_help();
		return 0xff;
	}

	dump_path =iopt->argv[0];
	source_path =iopt->argv[1];

	if(!check_path(dump_path) || !check_path(source_path))
	{
		stegano_print_help();
		innova_p(FATAL, "invalid path passed as option");
	}

	return 0x00;
}
