/*
 * this file support select only, is called poll because
 * on the future there must implement the double support
 * select/epoll for system that support event poll device,
 * is a workaroud, must be expanded.
 *
 * 0.0.1:		simple select support
 *
 */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/select.h>

#include "innova.h"

/*
 * pointer to global innova struct, used for plugin argument
 */
extern struct innova_struct *current;

/*
 * for keep locally handle and file descriptor of the two peer
 */
static struct ul_handle_fd couple[2];

static fd_set r_set;
static int maxfd;

/*
 * this is lazy function, used only for a pseudo-common
 * interface usable for add stream to the queue of 
 * stream usable for input, on the future, with more 
 * socket type supported, a common interface should
 * be used, but calling a socket-dependent register 
 * function, keeping on commong, the FD_SET call only
 */
void add_innova_stream(struct ipulog_handle *stream, int fd)
{
	/* stream !=NULL -> local, stream ==NULL -> remote */
	if(stream !=NULL)
	{
		couple[LOCAL].ip_handle =(void *)stream;
		couple[LOCAL].fd =stream->fd;

		if(maxfd < stream->fd)
			maxfd =stream->fd;
	}
	else
	{
		couple[REMOTE].ip_handle =NULL;
		couple[REMOTE].fd =fd;

		/* netlink use local timeout and not NONBLOCK socket */
		if(maxfd < fd)
			maxfd =fd;
	}
}

/*
 * this function should be used for register file descriptor used
 * from plugin, simple is called and passed a function pointer, 
 * return on the argument innova struct and the file decriptor ready
 * for be readed.
 *
 * pfds is array for keep file descriptor
 * pfds_number is the max value of index on pfds and pcbs arrays
 * pcbs is plugin callbacks, function pointer array to call when
 * fd appears to be ready.
 */
static int pfds[MAX_PLUG_FDS];
static int pfds_number;
static void *pcbs[MAX_PLUG_FDS];

void add_fd(int fd, int (* callback)(struct innova_struct *, int))
{
	pfds[pfds_number] =fd;
	pcbs[pfds_number] =(void *)callback;

	pfds_number++;

	if(pfds_number == MAX_PLUG_FDS)
		innova_p(FATAL, "too much file descriptor, change MAX_PLUG_FDS innova.h");

	if(maxfd > fd)
		maxfd =fd;
}

void remove_fd(int fd)
{
	int i;

	for(i =0; i < pfds_number; i++)
	{
		/* 
		 * when is find, must be overwritten with 
		 * the next. but should be the last..
		 */
		if(pfds[i] ==fd)
		{
			while((i +1) < pfds_number)
			{
				pfds[i] =pfds[i+1];
				pcbs[i] =pcbs[i+1];

				i++;
			}
			pfds[i] =0x00;
			pcbs[i] =NULL;
		}
	}
}

struct innova_packet *wait_packet(int ms_timeout)
{
	/*
	 * static buffer for keep entry packets
	 */
	static unsigned char pkt_buf[BUFSIZE];
	/*
	 * packet_queued is used for sign is ulog library
	 * should have queued packets
	 */
	static int packet_queued;
	/* 
	 * statically declaration of the innova_packet
	 * struct, here referred, on get_ulog_packet and 
	 * get_datalink_packet 
	 */
	struct innova_packet *ret =NULL;
	struct timeval tv;

	/*
	 * callback useful for file descriptor add to select from the plugins,
	 * for this is not used the common local_mangle and remote_mangle
	 * how happens for datalink and netlink packets.
	 *
	 * the argument is innova_struct current and the file descriptor.
	 * from innova_struct all is reachable...
	 *
	 * return value is check and if is returned a value lower than
	 * 0, error is supposed (how for local_mangle and remote_mangle)
	 * and is returned NULL to handle_packets.
	 */
	int (* callback)(struct innova_struct *, int);

	/* this should be used with a lot of packets queued on netlink */
#if 0
	if(packet_queued)
	{
		ret =get_ulog_packet(couple[LOCAL].ip_handle, pkt_buf);

		if(ret ==NULL)
			packet_queued =0x00;
		/* 
		 * ret.diff is 0 and is ok! because the packet is on rapid 
		 * sequence
		 */
		else
			return ret;
	}
#endif

	FD_ZERO(&r_set);
	
	while(1)
	{
		int sret, i;

		FD_SET(couple[REMOTE].fd, &r_set);
		FD_SET(couple[LOCAL].fd, &r_set);

		for(i =0; i < pfds_number; i++)
			FD_SET(pfds[i], &r_set);
		
		tv.tv_sec =0;
		tv.tv_usec =ms_timeout;

		memset(pkt_buf, 0x00, BUFSIZE);

		sret =select(maxfd +1, &r_set, NULL, NULL, &tv);

		/* return NULL on handle_packets mean timeout expired */
		if(!sret)
			return NULL;

		if(sret ==-1)
			innova_p(FATAL_ERRNO, "select error");

		if(FD_ISSET(couple[LOCAL].fd, &r_set))
		{
			packet_queued =0xff;
			ret =get_ulog_packet(couple[LOCAL].ip_handle, pkt_buf);

			/* 
			 * reading from netlink should happens that NULL is returned,
			 * and here is required check or appears to be a timeout
			 * expired for handle_packets (innova.c)
			 */
			if(ret !=NULL)
			{
				ret->diff =(tv.tv_sec * 1000000) + tv.tv_usec;
				ret->diff =(ms_timeout - ret->diff);
				return ret;
			}
			continue;
		}

		if(FD_ISSET(couple[REMOTE].fd, &r_set))
		{
			ret =get_datalink_packet(couple[REMOTE].fd, pkt_buf);

			if(ret !=NULL)
			{
				ret->diff =(tv.tv_sec * 1000000) + tv.tv_usec;
				ret->diff =(ms_timeout - ret->diff);
				return ret;
			}
			continue;
		}

		/* 
		 * if plugin file descriptor list exist, check it and call 
		 * eventual call back function, when it return, we simple 
		 * back to the start of the while(1) loop, read other information
		 * on the head of the function and the file.
		 */
		for(i =0; i < pfds_number; i++)
		{
			if(FD_ISSET(pfds[i], &r_set))
			{
				(void *)callback =pcbs[i];

				if(callback(current, pfds[i]) < 0)
					return NULL;
			}
		}
	}
}
